import itertools
import math

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.tests.conftest import TORCH_AVAILABLE
from darts.utils.timeseries_generation import gaussian_timeseries
from darts.utils.utils import freqs

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )

from darts.utils.data import (
    HorizonBasedTorchTrainingDataset,
    SequentialTorchInferenceDataset,
    SequentialTorchTrainingDataset,
    ShiftedTorchTrainingDataset,
)


class TestDataset:
    target1 = gaussian_timeseries(length=100).with_static_covariates(
        pd.Series([0, 1], index=["st1", "st2"])
    )
    target2 = gaussian_timeseries(length=150).with_static_covariates(
        pd.Series([2, 3], index=["st1", "st2"])
    )
    cov_st1 = target1.static_covariates.values
    cov_st2 = target2.static_covariates.values
    cov_st2_df = pd.Series([2, 3], index=["st1", "st2"])
    vals1, vals2 = target1.values(), target2.values()
    cov1, cov2 = gaussian_timeseries(length=100), gaussian_timeseries(length=150)

    def _assert_eq_schema(self, left: dict, right: dict):
        assert all([
            (v == right[k] if not isinstance(v, pd.DataFrame) else v.equals(right[k]))
            for k, v in left.items()
        ])

    def _assert_eq(self, lefts: tuple, rights: tuple):
        for left, right in zip(lefts, rights):
            left = left.values() if isinstance(left, TimeSeries) else left
            right = right.values() if isinstance(right, TimeSeries) else right
            assert type(left) is type(right)
            assert (
                isinstance(
                    left, (TimeSeries, pd.Series, pd.DataFrame, np.ndarray, list, dict)
                )
                or left is None
            )
            if isinstance(left, (pd.Series, pd.DataFrame)):
                assert left.equals(right)
            elif isinstance(left, np.ndarray):
                np.testing.assert_array_equal(left, right)
            elif isinstance(left, list):
                assert left == right
            elif isinstance(left, dict):
                self._assert_eq_schema(left, right)
            else:
                assert right is None

    def _check_ds_stride(self, ds_regular, ds_stride, stride: int):
        """
        Every `stride`-th values in a dataset with stride=1 should be identical to the dataset stridden with `stride
        """
        # if the un-stridden length is a multiple of the stride
        if len(ds_regular) % stride == 0:
            assert len(ds_regular) == len(ds_stride) * stride

        for idx, batch_str in enumerate(ds_stride):
            for entry_s, entry_r in zip(batch_str, ds_regular[idx * stride]):
                if entry_s is not None and entry_r is not None:
                    np.testing.assert_almost_equal(entry_s, entry_r)
                else:
                    assert entry_s == entry_r

    def test_past_covariates_inference_dataset(self):
        # one target series
        ds = SequentialTorchInferenceDataset(
            series=self.target1, input_chunk_length=len(self.target1)
        )
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self._assert_eq(
            ds[0][1:], (None, None, None, None, self.cov_st1, self.target1.schema())
        )

        # two target series
        ds = SequentialTorchInferenceDataset(
            series=[self.target1, self.target2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self._assert_eq(
            ds[1][1:], (None, None, None, None, self.cov_st2, self.target2.schema())
        )

        # fail if covariates do not have same size
        with pytest.raises(ValueError) as exc:
            ds = SequentialTorchInferenceDataset(
                series=[self.target1, self.target2], past_covariates=[self.cov1]
            )
        assert str(exc.value) == (
            "The sequence of `past_covariates` must have the same length as the sequence of target `series`."
        )

        # with covariates
        ds = SequentialTorchInferenceDataset(
            series=[self.target1, self.target2],
            past_covariates=[self.cov1, self.cov2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        np.testing.assert_almost_equal(ds[1][1], self.cov2.values())
        self._assert_eq(
            ds[1][2:], (None, None, None, self.cov_st2, self.target2.schema())
        )  # no "future past" covariate here

        # more complex case with future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100101", end="20100820", freq="D"
        )  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        short_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=short_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if covariates are too short
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value) == (
            "For the given forecasting horizon `n=30`, the provided `past_covariates` at series sequence index "
            "`0` do not extend far enough into the future. As `n > output_chunk_length` the `past_covariates` "
            "must end at or after time step `2010-07-21 00:00:00`, whereas now the end is at time "
            "step `2010-07-01 00:00:00`."
        )

        # Should return correct values when covariates is long enough
        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=long_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-30])
        assert ds[0][3] is None
        assert ds[0][4] is None
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        covariate = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=covariate,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:40])
        assert ds[0][3] is None
        assert ds[0][4] is None
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

    def test_future_covariates_inference_dataset(self):
        # one target series
        ds = SequentialTorchInferenceDataset(
            series=self.target1, input_chunk_length=len(self.target1)
        )
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self._assert_eq(
            ds[0][1:], (None, None, None, None, self.cov_st1, self.target1.schema())
        )

        # two target series
        ds = SequentialTorchInferenceDataset(
            series=[self.target1, self.target2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self._assert_eq(
            ds[1][1:], (None, None, None, None, self.cov_st2, self.target2.schema())
        )

        # fail if covariates do not have same size
        with pytest.raises(ValueError) as exc:
            ds = SequentialTorchInferenceDataset(
                series=[self.target1, self.target2],
                future_covariates=[self.cov1],
            )
        assert str(exc.value) == (
            "The sequence of `future_covariates` must have the same length as the sequence of target `series`."
        )

        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100101", end="20100820", freq="D"
        )  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        short_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            future_covariates=short_cov,
            input_chunk_length=10,
            n=30,
        )

        # should fail if covariates are too short
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value) == (
            "For the given forecasting horizon `n=30`, the provided `future_covariates` at series sequence "
            "index `0` do not extend far enough into the future. As `n > output_chunk_length` the "
            "`future_covariates` must end at or after time step `2010-07-31 00:00:00`, whereas now "
            "the end is at time step `2010-07-01 00:00:00`."
        )

        # Should return correct values when covariates is long enough
        ds = SequentialTorchInferenceDataset(
            series=target,
            future_covariates=long_cov,
            input_chunk_length=10,
            n=30,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        assert ds[0][1] is None
        assert ds[0][2] is None
        np.testing.assert_almost_equal(ds[0][3], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][4], long_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        covariate = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            future_covariates=covariate,
            input_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        assert ds[0][1] is None
        assert ds[0][2] is None
        np.testing.assert_almost_equal(ds[0][3], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][4], covariate.values()[30:50])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

    def test_dual_covariates_inference_dataset(self):
        # one target series
        ds = SequentialTorchInferenceDataset(
            series=self.target1, input_chunk_length=len(self.target1)
        )
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self._assert_eq(
            ds[0][1:], (None, None, None, None, self.cov_st1, self.target1.schema())
        )

        # two target series
        ds = SequentialTorchInferenceDataset(
            series=[self.target1, self.target2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self._assert_eq(
            ds[1][1:], (None, None, None, None, self.cov_st2, self.target2.schema())
        )

        # fail if covariates do not have same size
        with pytest.raises(ValueError):
            ds = SequentialTorchInferenceDataset(
                series=[self.target1, self.target2],
                future_covariates=[self.cov1],
            )

        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100101", end="20100820", freq="D"
        )  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        short_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            future_covariates=short_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = SequentialTorchInferenceDataset(
            series=target,
            future_covariates=long_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        assert ds[0][1] is None
        assert ds[0][2] is None
        np.testing.assert_almost_equal(ds[0][3], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][4], long_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        covariate = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            future_covariates=covariate,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        assert ds[0][1] is None
        assert ds[0][2] is None
        np.testing.assert_almost_equal(ds[0][3], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][4], covariate.values()[30:50])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

    def test_mixed_covariates_inference_dataset(self):
        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100201", end="20100820", freq="D"
        )  # ends 50 days after times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_past_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )
        future_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=past_cov,
            future_covariates=past_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if future covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=long_past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # It should contain:
        # past_target, past_covariates, future_past_covariates, historic_future_covariates, future_covariates
        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_past_cov.values()[-50:-30])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][4], future_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )
        future_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70)
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], past_cov.values()[30:40])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[10:20])
        np.testing.assert_almost_equal(ds[0][4], future_cov.values()[20:40])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

    def test_split_covariates_inference_dataset(self):
        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100201", end="20100820", freq="D"
        )  # ends 50 days after times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_past_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )
        future_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=past_cov,
            future_covariates=past_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if future covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=long_past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # It should contain:
        # past_target, past_covariates, future_past_covariates, historic_future_covariates,
        # future_covariates, future_past_covariates
        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_past_cov.values()[-50:-30])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][4], future_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )
        future_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70)
        )

        ds = SequentialTorchInferenceDataset(
            series=target,
            past_covariates=past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], past_cov.values()[30:40])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[10:20])
        np.testing.assert_almost_equal(ds[0][4], future_cov.values()[20:40])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        self._assert_eq_schema(ds[0][6], target.schema())

    @pytest.mark.parametrize(
        "config",
        [
            # (dataset class, whether contains future, future batch index)
            (SequentialTorchInferenceDataset, [], None),
            (SequentialTorchInferenceDataset, ["past"], None),
            (SequentialTorchInferenceDataset, ["future"], 4),
            (SequentialTorchInferenceDataset, ["past", "future"], 4),
        ],
    )
    def test_inference_dataset_output_chunk_shift(self, config):
        ds_cls, use_covs, future_idx = config
        ocl = 1
        ocs = 2
        target = self.target1[: -(ocl + ocs)]

        ds_covs = {}
        for cov_type in use_covs:
            ds_covs[cov_type + "_covariates"] = self.cov1

        with pytest.raises(ValueError) as err:
            _ = ds_cls(
                series=target,
                input_chunk_length=1,
                output_chunk_length=1,
                output_chunk_shift=1,
                n=2,
                **ds_covs,
            )
        assert str(err.value).startswith("Cannot perform auto-regression")

        # regular dataset with output shift=0 and ocl=3: the 3rd future values should be identical to the 1st future
        # values of a dataset with output shift=2 and ocl=1
        ds_reg = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            output_chunk_shift=0,
            n=1,
            **ds_covs,
        )

        ds_shift = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=1,
            output_chunk_shift=ocs,
            n=1,
            **ds_covs,
        )

        batch_reg, batch_shift = ds_reg[0], ds_shift[0]

        # shifted prediction starts 2 steps after regular prediction
        assert batch_reg[-1] == batch_shift[-1] - ocs * target.freq

        if future_idx is not None:
            # 3rd future values of regular ds must be identical to the 1st future values of shifted dataset
            np.testing.assert_array_equal(
                batch_reg[future_idx][ocs:], batch_shift[future_idx]
            )
            batch_reg = batch_reg[:future_idx] + batch_reg[future_idx + 1 :]
            batch_shift = batch_shift[:future_idx] + batch_shift[future_idx + 1 :]

        # without future part, the input will be identical between regular, and shifted dataset
        assert all([
            np.all(el_reg == el_shift)
            for el_reg, el_shift in zip(batch_reg[:-1], batch_shift[:-1])
        ])

    def test_inference_dataset_bounds(self):
        # target1 has length 100
        assert len(self.target1) == 100

        kwargs = {
            "input_chunk_length": 3,
            "output_chunk_length": 1,
            "n": 1,
        }

        # missing stride
        with pytest.raises(ValueError) as exc:
            SequentialTorchInferenceDataset(
                series=self.target1, stride=0, bounds=np.array([[3, 100]]), **kwargs
            )
        assert (
            str(exc.value)
            == "Must supply either both `stride` and `bounds`, or none of them."
        )

        # stride = 1
        ds = SequentialTorchInferenceDataset(
            series=self.target1, stride=1, bounds=np.array([[3, 100]]), **kwargs
        )
        # length 98
        assert len(ds) == 100 - 3 + 1
        # first two sample are from beginning of the target with stride 1
        np.testing.assert_array_almost_equal(ds[0][0], self.target1.values()[:3])
        self._assert_eq_schema(ds[0][-2], self.target1.schema())
        assert ds[0][-1] == self.target1._time_index[3]

        np.testing.assert_array_almost_equal(ds[1][0], self.target1.values()[1:4])
        self._assert_eq_schema(ds[1][-2], self.target1.schema())
        assert ds[1][-1] == self.target1._time_index[4]

        # last two sample are from end of the target with stride 1
        np.testing.assert_array_almost_equal(ds[96][0], self.target1.values()[-4:-1])
        self._assert_eq_schema(ds[96][-2], self.target1.schema())
        assert ds[96][-1] == self.target1._time_index[-1]
        np.testing.assert_array_almost_equal(ds[97][0], self.target1.values()[-3:])
        self._assert_eq_schema(ds[97][-2], self.target1.schema())
        assert ds[97][-1] == self.target1._time_index[-1] + self.target1.freq

        # stride = 2, setting bounds upper limit as `100` can still only compute until `99` since starting
        # at `3` with stride
        ds = SequentialTorchInferenceDataset(
            series=self.target1, stride=2, bounds=np.array([[3, 100]]), **kwargs
        )

        # length 49
        assert len(ds) == math.ceil((100 - 3 + 1) / 2)
        # first two sample are from beginning of the target
        np.testing.assert_array_almost_equal(ds[0][0], self.target1.values()[:3])
        assert ds[0][-1] == self.target1._time_index[3]
        np.testing.assert_array_almost_equal(ds[1][0], self.target1.values()[2:5])
        assert ds[1][-1] == self.target1._time_index[5]
        # last two sample are from end of the target
        np.testing.assert_array_almost_equal(ds[47][0], self.target1.values()[-6:-3])
        assert ds[47][-1] == self.target1._time_index[-3]
        np.testing.assert_array_almost_equal(ds[48][0], self.target1.values()[-4:-1])
        assert ds[48][-1] == self.target1._time_index[-1]

        # stride = 2, output_chunk_shift = 1, same past target values but pred time is shifted by `+1`
        ds = SequentialTorchInferenceDataset(
            series=self.target1,
            stride=2,
            output_chunk_shift=1,
            bounds=np.array([[3, 100]]),
            **kwargs,
        )

        # length 49
        assert len(ds) == math.ceil((100 - 3 + 1 - 1) / 2)
        np.testing.assert_array_almost_equal(ds[0][0], self.target1.values()[:3])
        assert ds[0][-1] == self.target1._time_index[4]
        np.testing.assert_array_almost_equal(ds[1][0], self.target1.values()[2:5])
        assert ds[1][-1] == self.target1._time_index[6]
        np.testing.assert_array_almost_equal(ds[47][0], self.target1.values()[-6:-3])
        assert ds[47][-1] == self.target1._time_index[-2]
        np.testing.assert_array_almost_equal(ds[48][0], self.target1.values()[-4:-1])
        assert ds[48][-1] == self.target1._time_index[-1] + self.target1.freq

        # stride = 2, setting bounds upper limit as `101` will result in an index error for sample 50
        ds = SequentialTorchInferenceDataset(
            series=self.target1, stride=2, bounds=np.array([[3, 101]]), **kwargs
        )

        # length 50
        assert len(ds) == math.ceil((101 - 3 + 1) / 2)
        # getting the samples from before works
        np.testing.assert_array_almost_equal(ds[0][0], self.target1.values()[:3])
        assert ds[0][-1] == self.target1._time_index[3]
        np.testing.assert_array_almost_equal(ds[1][0], self.target1.values()[2:5])
        assert ds[1][-1] == self.target1._time_index[5]
        np.testing.assert_array_almost_equal(ds[47][0], self.target1.values()[-6:-3])
        assert ds[47][-1] == self.target1._time_index[-3]
        np.testing.assert_array_almost_equal(ds[48][0], self.target1.values()[-4:-1])
        assert ds[48][-1] == self.target1._time_index[-1]

        # but sample at index 50 raises an error
        with pytest.raises(IndexError):
            _ = ds[50]

    def test_inference_dataset_series_too_short(self):
        # stride = 2, setting bounds upper limit as `101` will result in an index error for sample 50
        ds = SequentialTorchInferenceDataset(
            series=self.target1, input_chunk_length=len(self.target1) + 1
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value).startswith(
            "The dataset contains target `series` that are too short"
        )

        # past covs start too late
        ds = SequentialTorchInferenceDataset(
            series=self.target1,
            past_covariates=self.target1[1:],
            input_chunk_length=len(self.target1),
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value).startswith(
            "For the given forecasting case, the provided `past_covariates` at "
            "series sequence index `0` do not extend far enough into the past."
        )

        # past covs end too early
        ds = SequentialTorchInferenceDataset(
            series=self.target1,
            past_covariates=self.target1[:-1],
            input_chunk_length=len(self.target1),
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value).startswith(
            "For the given forecasting horizon `n=1`, the provided `past_covariates` at "
            "series sequence index `0` do not extend far enough into the future."
        )

        # past covs start too late
        target_short = self.target1[:-1]
        ds = SequentialTorchInferenceDataset(
            series=target_short,
            future_covariates=self.target1[1:],
            input_chunk_length=len(target_short),
            output_chunk_length=1,
            n=1,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value).startswith(
            "For the given forecasting case, the provided `future_covariates` at "
            "series sequence index `0` do not extend far enough into the past."
        )

        # future covs end too early
        ds = SequentialTorchInferenceDataset(
            series=target_short,
            future_covariates=target_short,
            input_chunk_length=len(target_short),
            output_chunk_length=1,
            n=1,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value).startswith(
            "For the given forecasting horizon `n=1`, the provided `future_covariates` at "
            "series sequence index `0` do not extend far enough into the future."
        )

    def test_shifted_training_dataset_too_short(self):
        # one target series
        with pytest.raises(ValueError) as exc:
            _ = ShiftedTorchTrainingDataset(
                series=self.target1[:5],
                input_chunk_length=3,
                output_chunk_length=3,
                shift=3,
            )
        assert str(exc.value) == (
            "The input `series` are too short to extract even a single sample. "
            "Expected min length: `6`, received max length: `5`."
        )

        # two target series both too short, will hint at max length of both
        with pytest.raises(ValueError) as exc:
            _ = ShiftedTorchTrainingDataset(
                series=[self.target1[:3], self.target1[:4]],
                input_chunk_length=3,
                output_chunk_length=3,
                shift=3,
            )
        assert str(exc.value) == (
            "The input `series` are too short to extract even a single sample. "
            "Expected min length: `6`, received max length: `4`."
        )

        # two target series, first is long enough, second is too short;
        # error is raised only when going through the dataset
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1[:6], self.target1[:5]],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=3,
        )
        # first sample of first series is okay
        _ = ds[0]
        # first sample of second series failed
        with pytest.raises(ValueError) as exc:
            _ = ds[1]
        assert str(exc.value) == (
            "The dataset contains target `series` that are too short to extract "
            "even a single example. Expected min length: `6`, received length `5` "
            "(at series sequence idx `1`)."
        )

    def test_horizon_training_dataset_too_short(self):
        # two target series, first is long enough, second is too short;
        # horizon based only detects too short series when going through the dataset
        ds = HorizonBasedTorchTrainingDataset(
            series=[self.target1[:6], self.target1[:5]],
            output_chunk_length=3,
            lookback=1,
            lh=(1, 1),
        )
        # first sample of first series is okay
        _ = ds[0]
        # first sample of second series failed
        with pytest.raises(ValueError) as exc:
            _ = ds[1]
        assert str(exc.value) == (
            "The dataset contains target `series` that are too short to extract "
            "even a single example. Expected min length: `6`, received length `5` "
            "(at series sequence idx `1`)."
        )
        # dataset end
        with pytest.raises(IndexError):
            _ = ds[2]

    def test_horizon_training_dataset_invalid_lh(self):
        # lh elements must be >= 1
        with pytest.raises(ValueError) as exc:
            _ = HorizonBasedTorchTrainingDataset(
                series=self.target1,
                output_chunk_length=3,
                lookback=1,
                lh=(1, 0),
            )
        assert str(exc.value) == (
            "Invalid `lh=(1, 0)`. `lh` must be a tuple `(min_lh, max_lh)`, "
            "with `1 <= min_lh <= max_lh`."
        )

    def test_past_covariates_sequential_dataset(self):
        # one target series
        ds = SequentialTorchTrainingDataset(
            series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 81
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )

        # two target series
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 262
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[136],
            (
                self.target2[125:135],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[125:135],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError) as exc:
            ds = SequentialTorchTrainingDataset(
                series=[self.target1, self.target2], past_covariates=[self.cov1]
            )
        assert str(exc.value) == (
            "The sequence of `past_covariates` must have the same length as the sequence of target `series`."
        )

        # two targets and two covariates
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            past_covariates=[self.cov1, self.cov2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                self.cov1[75:85],
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[136],
            (
                self.target2[125:135],
                self.cov2[125:135],
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # should fail if covariates do not have the required time span, even though covariates are longer
        times1 = pd.date_range(start="20100101", end="20110101", freq="D")
        times2 = pd.date_range(start="20120101", end="20150101", freq="D")
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = SequentialTorchTrainingDataset(
            series=target,
            past_covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[5]

        assert str(exc.value) == (
            "Invalid `past_covariates`; could not find values in index range: "
            "2010-12-08 00:00:00 - 2010-12-17 00:00:00."
        )

        # the same should fail when series are integer-indexed
        times1 = pd.RangeIndex(start=0, stop=100, step=1)
        times2 = pd.RangeIndex(start=200, stop=400, step=1)
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = SequentialTorchTrainingDataset(
            series=target,
            past_covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[5]
        assert str(exc.value) == (
            "Invalid `past_covariates`; could not find values in index range: 75 - 84."
        )

        # we should get the correct covariate slice even when target and covariates are not aligned
        times1 = pd.date_range(start="20100101", end="20110101", freq="D")
        times2 = pd.date_range(start="20090101", end="20110106", freq="D")
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = SequentialTorchTrainingDataset(
            series=target,
            past_covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-20:-10])
        np.testing.assert_almost_equal(ds[0][1], cov.values()[-25:-15])
        assert ds[0][2] is None  # historic future cov
        assert ds[0][3] is None  # future cov
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None  # sample weight
        np.testing.assert_almost_equal(ds[0][6], target.values()[-10:])

        np.testing.assert_almost_equal(ds[5][0], target.values()[-25:-15])
        np.testing.assert_almost_equal(ds[5][1], cov.values()[-30:-20])
        assert ds[5][2] is None  # historic future cov
        assert ds[5][3] is None  # future cov
        np.testing.assert_almost_equal(ds[5][4], self.cov_st2)
        assert ds[5][5] is None  # sample weight
        np.testing.assert_almost_equal(ds[5][6], target.values()[-15:-5])

        # This should also be the case when series are integer indexed
        times1 = pd.RangeIndex(start=100, stop=200, step=1)
        times2 = pd.RangeIndex(start=50, stop=250, step=1)
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = SequentialTorchTrainingDataset(
            series=target,
            past_covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-20:-10])
        np.testing.assert_almost_equal(ds[0][1], cov.values()[-70:-60])
        assert ds[0][2] is None  # historic future cov
        assert ds[0][3] is None  # future cov
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None  # sample weight
        np.testing.assert_almost_equal(ds[0][6], target.values()[-10:])

        np.testing.assert_almost_equal(ds[5][0], target.values()[-25:-15])
        np.testing.assert_almost_equal(ds[5][1], cov.values()[-75:-65])
        assert ds[5][2] is None  # historic future cov
        assert ds[5][3] is None  # future cov
        np.testing.assert_almost_equal(ds[5][4], self.cov_st2)
        assert ds[5][5] is None  # sample weight
        np.testing.assert_almost_equal(ds[5][6], target.values()[-15:-5])

    def test_future_covariates_sequential_dataset(self):
        # one target series
        ds = SequentialTorchTrainingDataset(
            series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 81
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )

        # two target series
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 262
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[136],
            (
                self.target2[125:135],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[125:135],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError) as exc:
            ds = SequentialTorchTrainingDataset(
                series=[self.target1, self.target2],
                future_covariates=[self.cov1],
            )
        assert str(exc.value) == (
            "The sequence of `future_covariates` must have the same length as the sequence of target `series`."
        )

        # two targets and two covariates; covariates not aligned, must contain correct values
        target1 = TimeSeries.from_values(np.random.randn(100)).with_static_covariates(
            self.cov_st2_df
        )
        target2 = TimeSeries.from_values(np.random.randn(50)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(120))
        cov2 = TimeSeries.from_values(np.random.randn(80))

        ds = SequentialTorchTrainingDataset(
            series=[target1, target2],
            future_covariates=[cov1, cov2],
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-20:-10])
        assert ds[0][1] is None  # past cov
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-40:-30])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-30:-20])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None  # sample weight
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-10:])

        np.testing.assert_almost_equal(ds[101][0], target2.values()[-40:-30])
        assert ds[0][1] is None  # past cov
        np.testing.assert_almost_equal(ds[101][2], cov2.values()[-70:-60])
        np.testing.assert_almost_equal(ds[101][3], cov2.values()[-60:-50])
        np.testing.assert_almost_equal(ds[101][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[101][6], target2.values()[-30:-20])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = SequentialTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-4:-2])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-6:-4])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-4:-2])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-2:])

        # Should fail if covariates are not long enough
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))

        ds = SequentialTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert (
            str(exc.value)
            == "Invalid `future_covariates`; could not find values in index range: 6 - 7."
        )

    def test_dual_covariates_sequential_dataset(self):
        # Must contain (past_target, historic_future_covariates, future_covariates, static covariates,
        # sample weight, future_target)

        # one target series
        ds = SequentialTorchTrainingDataset(
            series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 81
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )

        # two target series
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 262
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[136],
            (
                self.target2[125:135],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = SequentialTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[125:135],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = SequentialTorchTrainingDataset(
                series=[self.target1, self.target2],
                future_covariates=[self.cov1],
            )

        # two targets and two covariates; covariates not aligned, must contain correct values
        target1 = TimeSeries.from_values(np.random.randn(100)).with_static_covariates(
            self.cov_st2_df
        )
        target2 = TimeSeries.from_values(np.random.randn(50)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(120))
        cov2 = TimeSeries.from_values(np.random.randn(80))

        ds = SequentialTorchTrainingDataset(
            series=[target1, target2],
            future_covariates=[cov1, cov2],
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-20:-10])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-40:-30])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-30:-20])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-10:])

        np.testing.assert_almost_equal(ds[101][0], target2.values()[-40:-30])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[101][2], cov2.values()[-70:-60])
        np.testing.assert_almost_equal(ds[101][3], cov2.values()[-60:-50])
        np.testing.assert_almost_equal(ds[101][4], self.cov_st2)
        assert ds[101][5] is None
        np.testing.assert_almost_equal(ds[101][6], target2.values()[-30:-20])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = SequentialTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-4:-2])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-6:-4])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-4:-2])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-2:])

        # Should fail if covariates are not long enough
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))

        ds = SequentialTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        with pytest.raises(ValueError):
            _ = ds[0]

    def test_past_covariates_shifted_dataset(self):
        # one target series
        ds = ShiftedTorchTrainingDataset(
            series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        assert len(ds) == 86
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )

        # two target series
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        assert len(ds) == 272
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[130:140],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = ShiftedTorchTrainingDataset(
                series=[self.target1, self.target2], past_covariates=[self.cov1]
            )

        # covariates end too early
        chunk_length = 3
        series = self.target1[: 2 * chunk_length]
        ds = ShiftedTorchTrainingDataset(
            series=series,
            past_covariates=series[: -(chunk_length + 1)],
            input_chunk_length=chunk_length,
            output_chunk_length=chunk_length,
            shift=chunk_length,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value) == (
            "Invalid `past_covariates`; could not find values in index range: "
            "2000-01-01 00:00:00 - 2000-01-03 00:00:00."
        )

        # covariates are long enough but don't have the same frequency
        ds = ShiftedTorchTrainingDataset(
            series=series,
            past_covariates=self.target1[::2],
            input_chunk_length=chunk_length,
            output_chunk_length=chunk_length,
            shift=chunk_length,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value) == (
            "The `past_covariates` frequency `<2 * Days>` does not match "
            "the target `series` frequency `<Day>` (at series sequence idx `0`)."
        )

        # two targets and two covariates
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            past_covariates=[self.cov1, self.cov2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                self.cov1[80:90],
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                self.cov2[130:140],
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            past_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
        assert ds[0][2] is None
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-3:])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            past_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
        assert ds[0][2] is None
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-3:])

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(5))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            past_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        with pytest.raises(ValueError):
            _ = ds[0]

    def test_future_covariates_shifted_dataset(self):
        # one target series
        ds = ShiftedTorchTrainingDataset(
            series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        assert len(ds) == 86
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )

        # two target series
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        assert len(ds) == 272
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[130:140],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = ShiftedTorchTrainingDataset(
                series=[self.target1, self.target2],
                future_covariates=[self.cov1],
            )

        # covariates end too early
        chunk_length = 3
        series = self.target1[: 2 * chunk_length]
        ds = ShiftedTorchTrainingDataset(
            series=series,
            future_covariates=series[:-1],
            input_chunk_length=chunk_length,
            output_chunk_length=chunk_length,
            shift=chunk_length,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value) == (
            "Invalid `future_covariates`; could not find values in index range: "
            "2000-01-04 00:00:00 - 2000-01-06 00:00:00."
        )

        # covariates are long enough but don't have the same frequency
        ds = ShiftedTorchTrainingDataset(
            series=series,
            future_covariates=self.target1[::2],
            input_chunk_length=chunk_length,
            output_chunk_length=chunk_length,
            shift=chunk_length,
        )
        with pytest.raises(ValueError) as exc:
            _ = ds[0]
        assert str(exc.value) == (
            "The `future_covariates` frequency `<2 * Days>` does not match "
            "the target `series` frequency `<Day>` (at series sequence idx `0`)."
        )

        # two targets and two covariates
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            future_covariates=[self.cov1, self.cov2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                self.cov1[80:90],
                self.cov1[85:95],
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                None,
                self.cov2[130:140],
                self.cov2[135:145],
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-3:])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-3:])

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        with pytest.raises(ValueError):
            _ = ds[0]

    def test_dual_covariates_shifted_dataset(self):
        # one target series
        ds = ShiftedTorchTrainingDataset(
            series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        assert len(ds) == 86
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )

        # two target series
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        assert len(ds) == 272
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                None,
                None,
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[130:140],
                None,
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = ShiftedTorchTrainingDataset(
                series=[self.target1, self.target2],
                future_covariates=[self.cov1],
            )

        # two targets and two covariates
        ds = ShiftedTorchTrainingDataset(
            series=[self.target1, self.target2],
            future_covariates=[self.cov1, self.cov2],
            input_chunk_length=10,
            output_chunk_length=10,
            shift=5,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                None,
                self.cov1[80:90],
                self.cov1[85:95],
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                None,
                self.cov2[130:140],
                self.cov2[135:145],
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-3:])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        assert ds[0][1] is None
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][3], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] is None
        np.testing.assert_almost_equal(ds[0][6], target1.values()[-3:])

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))
        ds = ShiftedTorchTrainingDataset(
            series=[target1],
            future_covariates=[cov1],
            input_chunk_length=3,
            output_chunk_length=3,
            shift=2,
        )
        with pytest.raises(ValueError):
            _ = ds[0]

    @pytest.mark.parametrize("use_weight", [False, True])
    def test_horizon_based_dataset(self, use_weight):
        ds_kwargs = {
            "output_chunk_length": 10,
            "lh": (1, 3),
            "lookback": 2,
        }
        weight1 = self.target1 + 1
        weight2 = self.target2 + 1

        weight_exp = weight1[85:95] if use_weight else None
        # one target series
        ds_kwargs["series"] = self.target1
        ds_kwargs["sample_weight"] = weight1 if use_weight else None
        ds = HorizonBasedTorchTrainingDataset(**ds_kwargs)
        # 21 as both `lh` bounds are inclusive
        assert len(ds) == 21
        self._assert_eq(
            ds[5],
            (
                self.target1[65:85],
                None,
                None,
                None,
                self.cov_st1,
                weight_exp,
                self.target1[85:95],
            ),
        )
        # one target series, with stride
        ds_stride = HorizonBasedTorchTrainingDataset(**ds_kwargs, stride=3)
        self._check_ds_stride(ds_regular=ds, ds_stride=ds_stride, stride=3)

        # two target series
        weight_exp1 = weight1[85:95] if use_weight else None
        weight_exp2 = weight2[135:145] if use_weight else None
        ds_kwargs["series"] = [self.target1, self.target2]
        ds_kwargs["sample_weight"] = [weight1, weight2] if use_weight else None
        ds = HorizonBasedTorchTrainingDataset(**ds_kwargs)
        # 42 as both `lh` bounds are inclusive per series
        assert len(ds) == 42
        self._assert_eq(
            ds[5],
            (
                self.target1[65:85],
                None,
                None,
                None,
                self.cov_st1,
                weight_exp1,
                self.target1[85:95],
            ),
        )
        # 21 samples after comes the second series
        self._assert_eq(
            ds[26],
            (
                self.target2[115:135],
                None,
                None,
                None,
                self.cov_st2,
                weight_exp2,
                self.target2[135:145],
            ),
        )
        # two target series, with stride
        ds_stride = HorizonBasedTorchTrainingDataset(**ds_kwargs, stride=3)
        self._check_ds_stride(ds_regular=ds, ds_stride=ds_stride, stride=3)

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = HorizonBasedTorchTrainingDataset(
                series=[self.target1, self.target2], past_covariates=[self.cov1]
            )

        # two targets and two covariates
        weight_exp1 = weight1[85:95] if use_weight else None
        weight_exp2 = weight2[135:145] if use_weight else None
        ds_kwargs["series"] = [self.target1, self.target2]
        ds_kwargs["past_covariates"] = [self.cov1, self.cov2]
        ds_kwargs["sample_weight"] = [weight1, weight2] if use_weight else None
        ds = HorizonBasedTorchTrainingDataset(**ds_kwargs)
        self._assert_eq(
            ds[5],
            (
                self.target1[65:85],
                self.cov1[65:85],
                None,
                None,
                self.cov_st1,
                weight_exp1,
                self.target1[85:95],
            ),
        )
        # 21 samples after comes the second series
        self._assert_eq(
            ds[26],
            (
                self.target2[115:135],
                self.cov2[115:135],
                None,
                None,
                self.cov_st2,
                weight_exp2,
                self.target2[135:145],
            ),
        )
        # two targets and two covariates, with stride
        ds_stride = HorizonBasedTorchTrainingDataset(**ds_kwargs, stride=3)
        self._check_ds_stride(ds_regular=ds, ds_stride=ds_stride, stride=3)

    @pytest.mark.parametrize(
        "config",
        [
            # (dataset class, whether contains future, future batch index)
            (SequentialTorchTrainingDataset, [], None),
            (SequentialTorchTrainingDataset, ["past"], None),
            (SequentialTorchTrainingDataset, ["future"], 3),
            (SequentialTorchTrainingDataset, ["past", "future"], 3),
        ],
    )
    def test_sequential_training_dataset_output_chunk_shift(self, config):
        ds_cls, use_covs, future_idx = config
        ocl = 1
        ocs = 2
        target = self.target1[: -(ocl + ocs)]
        sample_weight = target + 1

        ds_covs = {}
        for cov_type in use_covs:
            ds_covs[cov_type + "_covariates"] = self.cov1

        # regular dataset with output shift=0 and ocl=3: the 3rd future values should be identical to the 1st future
        # values of a dataset with output shift=2 and ocl=1
        ds_reg = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            output_chunk_shift=0,
            sample_weight=sample_weight,
            **ds_covs,
        )

        ds_shift = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=1,
            output_chunk_shift=ocs,
            sample_weight=sample_weight,
            **ds_covs,
        )

        batch_reg, batch_shift = ds_reg[0], ds_shift[0]

        if future_idx is not None:
            # 3rd future values of regular ds must be identical to the 1st future values of shifted dataset
            np.testing.assert_array_equal(
                batch_reg[future_idx][-1:], batch_shift[future_idx]
            )
            batch_reg = batch_reg[:future_idx] + batch_reg[future_idx + 1 :]
            batch_shift = batch_shift[:future_idx] + batch_shift[future_idx + 1 :]

        # last two elements are (sample weight, output chunk of the target series).
        # 3rd future values of regular ds must be identical to the 1st future values of shifted dataset
        batch_reg = batch_reg[:-2] + (batch_reg[-2][ocs:], batch_reg[-1][ocs:])

        # without future part, the input will be identical between regular, and shifted dataset
        assert all([
            np.all(el_reg == el_shift)
            for el_reg, el_shift in zip(batch_reg[:-1], batch_shift[:-1])
        ])

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [
                (SequentialTorchTrainingDataset, []),
                (SequentialTorchTrainingDataset, ["past"]),
                (SequentialTorchTrainingDataset, ["future"]),
                (SequentialTorchTrainingDataset, ["past", "future"]),
            ],
            [True, False],
        ),
    )
    def test_sequential_training_dataset_weight(self, config):
        (ds_cls, use_covs), manual_weight = config

        def get_built_in_weigths(targets):
            if isinstance(targets, list):
                max_steps = max([len(ts) for ts in targets])
            else:
                max_steps = len(targets)
            weight_expected = np.linspace(0, 1, max_steps)[-3:]
            return np.expand_dims(weight_expected, -1)

        target1 = self.target1
        target2 = self.target2
        weight1 = target1 + 1
        weight2 = target2 + 1
        built_in_weight = "linear"

        ds_covs = {}
        for cov_type in use_covs:
            ds_covs[cov_type + "_covariates"] = self.cov1

        # no sample weight
        ds = ds_cls(
            series=target1,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=None,
            **ds_covs,
        )
        assert ds[0][-2] is None

        # whenever we use sample weight, the weight are extracted from the same time frame as the target labels
        # since we set the weight to be `target + 1`, the returned batch weight must also be `batch_target_label + 1`

        # single univariate
        target = target1
        weight = weight1 if manual_weight else built_in_weight
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # single univariate with longer weight
        target = target1
        weight = (
            weight1.prepend_values([0.0]).append_values([0.0])
            if manual_weight
            else built_in_weight
        )
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # single multivariate with multivariate weight
        target = target1.stack(target1 + 1)
        weight = weight1.stack(weight1 + 1) if manual_weight else built_in_weight
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # single multivariate with univariate (global) weight
        target = target1.stack(target1 + 1)
        weight = weight1 if manual_weight else built_in_weight
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        # output weight corresponds to first target component + 1 (e.g. weight1)
        weight_exp = (
            ds[0][-1][:, 0:1] + 1 if manual_weight else get_built_in_weigths(target)
        )
        assert np.all(ds[0][-2] == weight_exp)

        # single univariate and list of single weight
        target = target1
        weight = [weight1] if manual_weight else built_in_weight
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # multiple univariate
        target = [target1, target2]
        weight = [weight1, weight2] if manual_weight else built_in_weight
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **{k: [v] * 2 for k, v in ds_covs.items()},
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # multiple multivariate
        target = [target1.stack(target1 + 1), target2.stack(target2 + 1)]
        weight = (
            [weight1.stack(weight1 + 1), weight2.stack(weight2 + 1)]
            if manual_weight
            else built_in_weight
        )
        ds = ds_cls(
            series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **{k: [v] * 2 for k, v in ds_covs.items()},
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

    def test_sequential_training_dataset_invalid_weight(self):
        ds_cls = SequentialTorchTrainingDataset
        ts = self.target1

        # invalid built-in weight
        with pytest.raises(ValueError) as err:
            _ = ds_cls(
                series=[ts, ts],
                input_chunk_length=1,
                output_chunk_length=3,
                sample_weight="invalid",
            )
        assert str(err.value).startswith(
            "Invalid `sample_weight` value: `'invalid'`. If a string, must be one of: "
        )

        # mismatch number of target and weight series
        with pytest.raises(ValueError) as err:
            _ = ds_cls(
                series=[ts, ts],
                input_chunk_length=1,
                output_chunk_length=3,
                sample_weight=[ts],
            )
        assert (
            str(err.value)
            == "The provided sequence of target `series` must have the same "
            "length as the provided sequence of `sample_weight`."
        )

        # too many weight components
        ds = ds_cls(
            series=ts,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=ts.stack(ts + 1),
        )
        with pytest.raises(ValueError) as err:
            _ = ds[0]
        assert (
            str(err.value)
            == "The number of components in `sample_weight` must either be `1` or match "
            "the number of target series components `1` (at series sequence idx `0`)."
        )

        # weight too short end
        ds = ds_cls(
            series=ts,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=ts[:-1],
        )
        with pytest.raises(ValueError) as err:
            _ = ds[0]
        assert (
            str(err.value)
            == "Invalid `sample_weight`; could not find values in index range: "
            "2000-04-07 00:00:00 - 2000-04-09 00:00:00."
        )

        # weight too short start
        ds = ds_cls(
            series=ts,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=ts[2:],
        )
        with pytest.raises(ValueError) as err:
            _ = ds[len(ds) - 1]
        assert (
            str(err.value)
            == "Invalid `sample_weight`; could not find values in index range: "
            "2000-01-02 00:00:00 - 2000-01-04 00:00:00."
        )

    @pytest.mark.parametrize(
        "config",
        [
            (SequentialTorchTrainingDataset, []),
            (SequentialTorchTrainingDataset, ["past"]),
            (SequentialTorchTrainingDataset, ["future"]),
            (SequentialTorchTrainingDataset, ["past", "future"]),
        ],
    )
    def test_sequential_training_dataset_stride(self, config):
        ds_cls, use_covs = config

        ds_covs = {}
        for cov_type in use_covs:
            ds_covs[cov_type + "_covariates"] = self.cov1

        ds_cls, future_idx = config
        icl = 4
        ocl = 2
        nb_samples = 12
        target = self.target1[: icl + ocl + nb_samples - 1]

        ds_reg = ds_cls(
            series=target,
            input_chunk_length=icl,
            output_chunk_length=ocl,
            stride=1,
            **ds_covs,
        )

        ds_stride = ds_cls(
            series=target,
            input_chunk_length=icl,
            output_chunk_length=ocl,
            stride=3,
            **ds_covs,
        )
        assert len(ds_stride) * 3 == len(ds_reg) == nb_samples
        self._check_ds_stride(ds_regular=ds_reg, ds_stride=ds_stride, stride=3)

    def test_get_matching_index(self):
        from darts.utils.data.utils import _get_matching_index

        # Check dividable freq
        times1 = pd.date_range(start="20100101", end="20100330", freq="D")
        times2 = pd.date_range(start="20100101", end="20100320", freq="D")
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        assert _get_matching_index(target, cov, idx=15) == 5

        # check non-dividable freq
        times1 = pd.date_range(start="20100101", end="20120101", freq=freqs["ME"])
        times2 = pd.date_range(start="20090101", end="20110601", freq=freqs["ME"])
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        assert _get_matching_index(target, cov, idx=15) == 15 - 7

        # check integer-indexed series
        times2 = pd.RangeIndex(start=10, stop=90)
        target = TimeSeries.from_values(np.random.randn(100)).with_static_covariates(
            self.cov_st2_df
        )
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        assert _get_matching_index(target, cov, idx=15) == 5



from collections.abc import Sequence
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
import torch

# `TorchTrainingDataset` output
# (past target, past cov, historic future cov, future cov, static cov, sample weight, future target)
TorchTrainingDatasetOutput = tuple[
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    np.ndarray,
]
# `TorchTrainingDataset` output converted to batch with `torch.Tensor`
TorchTrainingBatch = tuple[
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    torch.Tensor,
]
# training sample has no sample weight
# (past target, past cov, historic future cov, future cov, static cov, future target)
TorchTrainingSample = tuple[
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    np.ndarray,
]


# `TorchInferenceDataset` output
# (past target, past cov, future past cov, historic future cov, future cov, static cov, target series schema, pred time)
TorchInferenceDatasetOutput = tuple[
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Optional[np.ndarray],
    dict[str, Any],
    Union[pd.Timestamp, int],
]
# `TorchInferenceDataset` output converted to batch with `torch.Tensor`
TorchInferenceBatch = tuple[
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Sequence[dict[str, Any]],
    Union[Sequence[pd.Timestamp], Sequence[int]],
]

# The useful batch features are
# (past target, past cov, historic future cov, future cov, static cov)
TorchBatch = tuple[
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
    Optional[torch.Tensor],
]

# the final module input is a tuple of three tensors where the past features concatenated
# (past features (past target + past cov + historic future cov), future cov, static cov)
PLModuleInput = tuple[torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]



from itertools import product
from typing import Union

import numpy as np
import pandas as pd
import pytest
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression

from darts import TimeSeries
from darts.metrics import mape, rmse
from darts.models import (
    LinearRegressionModel,
    NaiveDrift,
    NaiveSeasonal,
    RandomForestModel,
    RegressionEnsembleModel,
    SKLearnModel,
    Theta,
)
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.tests.models.forecasting.test_ensemble_models import _make_ts
from darts.tests.models.forecasting.test_sklearn_models import train_test_split
from darts.utils import timeseries_generation as tg

if TORCH_AVAILABLE:
    import torch

    from darts.models import BlockRNNModel, RNNModel


class TestRegressionEnsembleModels:
    RANDOM_SEED = 111

    sine_series = tg.sine_timeseries(
        value_frequency=(1 / 5), value_y_offset=10, length=50
    )
    lin_series = tg.linear_timeseries(length=50)

    combined = sine_series + lin_series

    seq1 = [_make_ts(0), _make_ts(10), _make_ts(20)]
    cov1 = [_make_ts(5), _make_ts(15), _make_ts(25)]

    seq2 = [_make_ts(0, 20), _make_ts(10, 20), _make_ts(20, 20)]
    cov2 = [_make_ts(5, 30), _make_ts(15, 30), _make_ts(25, 30)]

    # dummy feature and target TimeSeries instances
    ts_periodic = tg.sine_timeseries(length=500)
    ts_gaussian = tg.gaussian_timeseries(length=500)
    ts_random_walk = tg.random_walk_timeseries(length=500)

    ts_cov1 = ts_periodic.stack(ts_gaussian)
    ts_cov1 = ts_cov1.to_dataframe()
    ts_cov1.columns = ["Periodic", "Gaussian"]
    ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
    ts_sum1: TimeSeries = ts_periodic + ts_gaussian

    ts_cov2 = ts_sum1.stack(ts_random_walk)
    ts_sum2 = ts_sum1 + ts_random_walk

    def get_local_models(self):
        return [NaiveDrift(), NaiveSeasonal(5), NaiveSeasonal(10)]

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def get_global_models(
        self, output_chunk_length=5, input_chunk_length=20, training_length=24
    ):
        return [
            RNNModel(
                input_chunk_length=input_chunk_length,
                training_length=training_length,
                n_epochs=1,
                random_state=42,
                **tfm_kwargs,
            ),
            BlockRNNModel(
                input_chunk_length=input_chunk_length,
                output_chunk_length=output_chunk_length,
                n_epochs=1,
                random_state=42,
                **tfm_kwargs,
            ),
        ]

    @staticmethod
    def get_global_ensemble_model(output_chunk_length=5):
        lags = [-1, -2, -5]
        return RegressionEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
            ],
            regression_train_n_points=10,
        )

    def test_accepts_different_regression_models(self):
        regr1 = LinearRegression()
        regr2 = RandomForestRegressor()
        regr3 = RandomForestModel(lags_future_covariates=[0])

        model0 = RegressionEnsembleModel(self.get_local_models(), 10)
        model1 = RegressionEnsembleModel(self.get_local_models(), 10, regr1)
        model2 = RegressionEnsembleModel(self.get_local_models(), 10, regr2)
        model3 = RegressionEnsembleModel(self.get_local_models(), 10, regr3)

        models = [model0, model1, model2, model3]
        for model in models:
            model.fit(series=self.combined)
            model.predict(10)

    def test_accepts_one_model(self):
        regr1 = LinearRegression()
        regr2 = RandomForestModel(lags_future_covariates=[0])

        model0 = RegressionEnsembleModel([self.get_local_models()[0]], 10)
        model1 = RegressionEnsembleModel([self.get_local_models()[0]], 10, regr1)
        model2 = RegressionEnsembleModel([self.get_local_models()[0]], 10, regr2)

        models = [model0, model1, model2]
        for model in models:
            model.fit(series=self.combined)
            model.predict(10)

    def test_accept_pretrain_global_models(self):
        linreg1 = LinearRegressionModel(lags=1)
        linreg2 = LinearRegressionModel(lags=2)

        linreg1.fit(self.lin_series[:30])
        linreg2.fit(self.lin_series[:30])

        model_ens = RegressionEnsembleModel(
            forecasting_models=[linreg1, linreg2],
            regression_train_n_points=10,
            train_forecasting_models=False,
        )
        model_ens.fit(self.sine_series[:45])
        model_ens.predict(5)

        # train_forecasting_models=True requires all the model to be reset
        with pytest.raises(ValueError):
            RegressionEnsembleModel(
                forecasting_models=[linreg1, linreg2],
                regression_train_n_points=10,
                train_forecasting_models=True,
            )
        model_ens_ft = RegressionEnsembleModel(
            forecasting_models=[linreg1.untrained_model(), linreg2.untrained_model()],
            regression_train_n_points=10,
            train_forecasting_models=True,
        )
        model_ens_ft.fit(self.sine_series[:45])
        model_ens_ft.predict(5)

    def test_train_n_points(self):
        regr = LinearRegressionModel(lags_future_covariates=[0])

        # same values
        _ = RegressionEnsembleModel(self.get_local_models(), 5, regr)

        # too big value to perform the split
        ensemble = RegressionEnsembleModel(self.get_local_models(), 100)
        with pytest.raises(ValueError):
            ensemble.fit(self.combined)

        ensemble = RegressionEnsembleModel(self.get_local_models(), 50)
        with pytest.raises(ValueError):
            ensemble.fit(self.combined)

        # too big value considering min_train_series_length
        ensemble = RegressionEnsembleModel(self.get_local_models(), 45)
        with pytest.raises(ValueError):
            ensemble.fit(self.combined)

        # using regression_train_n_point=-1 without pretraining
        if TORCH_AVAILABLE:
            with pytest.raises(ValueError):
                RegressionEnsembleModel(
                    self.get_global_models(), regression_train_n_points=-1
                )

        # using regression_train_n_point=-1 with pretraining
        forecasting_models = [
            LinearRegressionModel(lags=1).fit(self.sine_series),
            LinearRegressionModel(lags=3).fit(self.sine_series),
        ]
        ensemble = RegressionEnsembleModel(
            forecasting_models=forecasting_models,
            regression_train_n_points=-1,
            train_forecasting_models=False,
        )
        ensemble.fit(self.combined)

        # 3 values are necessary to predict the first value for the 2nd forecasting model
        assert ensemble.ensemble_model.training_series == self.combined[3:]

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_torch_models_retrain(self):
        model1 = BlockRNNModel(
            input_chunk_length=12,
            output_chunk_length=1,
            random_state=0,
            n_epochs=2,
            **tfm_kwargs,
        )
        model2 = BlockRNNModel(
            input_chunk_length=12,
            output_chunk_length=1,
            random_state=0,
            n_epochs=2,
            **tfm_kwargs,
        )

        ensemble = RegressionEnsembleModel([model1], 5)
        # forecasting model is retrained from scratch with the entire series once the regression model is trained
        ensemble.fit(self.combined)
        model1_fitted = ensemble.forecasting_models[0]
        forecast1 = model1_fitted.predict(3)
        # train torch model outside of ensemble model
        model2.fit(self.combined)
        forecast2 = model2.predict(3)

        assert model1_fitted.training_series.time_index.equals(
            model2.training_series.time_index
        )
        assert forecast1.time_index.equals(forecast2.time_index)
        np.testing.assert_array_almost_equal(forecast1.values(), forecast2.values())

    @pytest.mark.parametrize("config", [(1, 1), (5, 2), (4, 3)])
    def test_train_with_historical_forecasts_no_covs(self, config):
        """
        Training regression model of ensemble with output from historical forecasts instead of predict should
        yield better results when the forecasting models are global and regression_train_n_points >> ocl.

        config[0] : both ocl = 1
        config[1] : both ocl are multiple of regression_train_n_points
        config[2] : ocl1 is multiple, ocl2 is not but series is long enough to shift the historical forecats start
        """
        ocl1, ocl2 = config
        regression_train_n_points = 20
        train, val = self.combined.split_after(self.combined.time_index[-10])

        # using predict to generate the future covs for the ensemble model
        ensemble_predict = RegressionEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(lags=5, output_chunk_length=ocl1),
                LinearRegressionModel(lags=2, output_chunk_length=ocl2),
            ],
            regression_train_n_points=regression_train_n_points,
            train_using_historical_forecasts=False,
        )
        ensemble_predict.fit(train)
        pred_predict = ensemble_predict.predict(len(val))

        assert (
            len(ensemble_predict.ensemble_model.training_series)
            == regression_train_n_points
        )

        # using historical forecasts to generate the future covs for the ensemble model
        ensemble_hist_fct = RegressionEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(lags=5, output_chunk_length=ocl1),
                LinearRegressionModel(lags=2, output_chunk_length=ocl2),
            ],
            regression_train_n_points=regression_train_n_points,
            train_using_historical_forecasts=True,
        )
        ensemble_hist_fct.fit(train)
        pred_hist_fct = ensemble_hist_fct.predict(len(val))

        assert (
            len(ensemble_hist_fct.ensemble_model.training_series)
            == regression_train_n_points
        )

        mape_hfc, mape_pred = mape(pred_hist_fct, val), mape(pred_predict, val)
        assert mape_hfc < mape_pred or mape_hfc == pytest.approx(mape_pred)

        rmse_hfc, rmse_pred = rmse(pred_hist_fct, val), rmse(pred_predict, val)
        assert rmse_hfc < rmse_pred or rmse_hfc == pytest.approx(rmse_pred)

    @pytest.mark.parametrize(
        "config",
        [(1, 1), (5, 2), (4, 3)],
    )
    def test_train_with_historical_forecasts_with_covs(self, config):
        """
        config[0] : both ocl = 1, covs are long enough
        config[1] : both ocl are multiple, covs are long enough
        config[2] : ocl1 multiple, ocl2 not multiple
        """
        ocl1, ocl2 = config
        regression_train_n_points = 10
        # shortening the series to make test simpler, 10 for forecasting models, 20 for the regression model
        ts = self.combined[:30]

        # past covariates starts 5 steps before the target series
        past_covs = tg.linear_timeseries(
            start=ts.start_time() - 5 * ts.freq,
            length=len(ts) + 5,
        )

        ensemble = RegressionEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(
                    lags=5, lags_past_covariates=5, output_chunk_length=ocl1
                ),
                LinearRegressionModel(
                    lags=2, lags_past_covariates=5, output_chunk_length=ocl2
                ),
            ],
            regression_train_n_points=regression_train_n_points,
            train_using_historical_forecasts=True,
        )
        # covariates have the appropriate length
        ensemble.fit(ts, past_covariates=past_covs)
        assert len(ensemble.ensemble_model.training_series) == regression_train_n_points
        # since past covariates extend far in the past, they are available for the regression model

        # future covariates finishes 5 steps after the target series
        future_covs = tg.linear_timeseries(
            start=ts.start_time(),
            length=len(ts) + 2,
        )

        ensemble = RegressionEnsembleModel(
            forecasting_models=[
                LinearRegressionModel(
                    lags=2, lags_future_covariates=[1, 2], output_chunk_length=ocl1
                ),
                LinearRegressionModel(
                    lags=1, lags_future_covariates=[1, 2], output_chunk_length=ocl2
                ),
            ],
            regression_train_n_points=regression_train_n_points,
            train_using_historical_forecasts=True,
        )

        # covariates have the appropriate length
        ensemble.fit(ts, future_covariates=future_covs)
        assert len(ensemble.ensemble_model.training_series) == regression_train_n_points

        with pytest.raises(ValueError):
            # covariates are too short (ends too early)
            ensemble.fit(ts, future_covariates=future_covs[: -min(ocl1, ocl2)])

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_train_predict_global_models_univar(self):
        ensemble_models = self.get_global_models(output_chunk_length=10)
        ensemble_models.append(SKLearnModel(lags=1))
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(series=self.combined)
        ensemble.predict(10)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_train_predict_global_models_multivar_no_covariates(self):
        ensemble_models = self.get_global_models(output_chunk_length=10)
        ensemble_models.append(SKLearnModel(lags=1))
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(self.seq1)
        ensemble.predict(10, self.seq1)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_train_predict_global_models_multivar_with_covariates(self):
        ensemble_models = self.get_global_models(output_chunk_length=10)
        ensemble_models.append(SKLearnModel(lags=1, lags_past_covariates=[-1]))
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(self.seq1, self.cov1)
        ensemble.predict(10, self.seq2, self.cov2)

    def test_train_predict_models_with_future_covariates(self):
        ensemble_models = [
            LinearRegressionModel(lags=1, lags_future_covariates=[1]),
            RandomForestModel(lags=1, lags_future_covariates=[1]),
        ]
        ensemble = RegressionEnsembleModel(ensemble_models, 10)
        ensemble.fit(self.sine_series, future_covariates=self.ts_cov1)
        # expected number of coefs is lags*components -> we have 1 lag for each target (1 comp)
        # and future covs (2 comp)
        expected_coefs = len(self.sine_series.components) + len(self.ts_cov1.components)
        assert len(ensemble.forecasting_models[0].model.coef_) == expected_coefs
        ensemble.predict(10, self.sine_series, future_covariates=self.ts_cov1)

    def test_predict_with_target(self):
        series_long = self.combined
        series_short = series_long[:25]

        # train with a single series
        ensemble_model = self.get_global_ensemble_model()
        ensemble_model.fit(series_short, past_covariates=series_long)
        # predict after end of train series
        preds = ensemble_model.predict(n=5, past_covariates=series_long)
        assert isinstance(preds, TimeSeries)
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        assert isinstance(preds, TimeSeries)
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        assert isinstance(preds, list) and len(preds) == 2
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        assert isinstance(preds, list) and len(preds) == 1

        # train with multiple series
        ensemble_model = self.get_global_ensemble_model()
        ensemble_model.fit([series_short] * 2, past_covariates=[series_long] * 2)
        with pytest.raises(ValueError):
            # predict without passing series should raise an error
            ensemble_model.predict(n=5, past_covariates=series_long)
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        assert isinstance(preds, TimeSeries)
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        assert isinstance(preds, list) and len(preds) == 2
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        assert isinstance(preds, list) and len(preds) == 1

    def helper_test_models_accuracy(
        self, model_instance, n, series, past_covariates, min_rmse
    ):
        # for every model, test whether it predicts the target with a minimum r2 score of `min_rmse`
        train_series, test_series = train_test_split(series, pd.Timestamp("20010101"))
        train_past_covariates, _ = train_test_split(
            past_covariates, pd.Timestamp("20010101")
        )

        model_instance.fit(series=train_series, past_covariates=train_past_covariates)
        prediction = model_instance.predict(n=n, past_covariates=past_covariates)
        current_rmse = rmse(test_series, prediction)

        assert current_rmse <= min_rmse, (
            f"Model was not able to denoise data. A rmse score of {current_rmse} was recorded."
        )

    def denoising_input(self):
        np.random.seed(self.RANDOM_SEED)

        ts_periodic = tg.sine_timeseries(length=500)
        ts_gaussian = tg.gaussian_timeseries(length=500)
        ts_random_walk = tg.random_walk_timeseries(length=500)

        ts_cov1 = ts_periodic.stack(ts_gaussian)
        ts_cov1 = ts_cov1.to_dataframe()
        ts_cov1.columns = ["Periodic", "Gaussian"]
        ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
        ts_sum1 = ts_periodic + ts_gaussian

        ts_cov2 = ts_sum1.stack(ts_random_walk)
        ts_sum2 = ts_sum1 + ts_random_walk

        return ts_sum1, ts_cov1, ts_sum2, ts_cov2

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_ensemble_models_denoising(self):
        # for every model, test whether it correctly denoises ts_sum using ts_gaussian and ts_sum as inputs
        # WARNING: this test isn't numerically stable, changing self.RANDOM_SEED can lead to exploding coefficients
        horizon = 10
        ts_sum1, ts_cov1, _, _ = self.denoising_input()
        torch.manual_seed(self.RANDOM_SEED)

        ensemble_models = [
            RNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
                **tfm_kwargs,
            ),
            BlockRNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
                **tfm_kwargs,
            ),
            SKLearnModel(lags_past_covariates=[-1]),
        ]

        ensemble = RegressionEnsembleModel(ensemble_models, horizon)
        self.helper_test_models_accuracy(ensemble, horizon, ts_sum1, ts_cov1, 3)

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    def test_ensemble_models_denoising_multi_input(self):
        # for every model, test whether it correctly denoises ts_sum_2 using ts_random_multi and ts_sum_2 as inputs
        # WARNING: this test isn't numerically stable, changing self.RANDOM_SEED can lead to exploding coefficients
        horizon = 10
        _, _, ts_sum2, ts_cov2 = self.denoising_input()
        torch.manual_seed(self.RANDOM_SEED)

        ensemble_models = [
            RNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
                **tfm_kwargs,
            ),
            BlockRNNModel(
                input_chunk_length=20,
                output_chunk_length=horizon,
                n_epochs=1,
                random_state=self.RANDOM_SEED,
                **tfm_kwargs,
            ),
            SKLearnModel(lags_past_covariates=[-1]),
            SKLearnModel(lags_past_covariates=[-1]),
        ]

        ensemble = RegressionEnsembleModel(ensemble_models, horizon)
        self.helper_test_models_accuracy(ensemble, horizon, ts_sum2, ts_cov2, 3)

    def test_call_backtest_regression_ensemble_local_models(self):
        regr_train_n = 10
        ensemble = RegressionEnsembleModel(
            forecasting_models=[NaiveSeasonal(5), Theta(2, 5)],
            regression_train_n_points=regr_train_n,
        )
        assert (
            max(m_.min_train_series_length for m_ in ensemble.forecasting_models) == 10
        )
        # 10 from training the forecasting models, and 10 from training the regression model
        assert ensemble.min_train_series_length == 10 + regr_train_n

        ensemble.fit(self.sine_series)
        # -10 comes from the maximum minimum train series length of all models
        assert ensemble.extreme_lags == (
            -10,
            -1,
            None,
            None,
            None,
            None,
            0,
        )
        preds = ensemble.historical_forecasts(self.sine_series)

        expected_start = (
            self.sine_series.start_time() + (10 + regr_train_n) * self.sine_series.freq
        )
        assert preds.start_time() == expected_start
        ensemble.backtest(self.sine_series)

    @pytest.mark.parametrize(
        "config",
        product(
            [
                (
                    2,
                    10,
                ),  # the 2 points are within first output chunk of m2 (4) -> min length=icl + ocl
                (4, 12),  # same as above
                (
                    5,
                    13,
                ),  # the fifth point is outside the first output chunk of m2 (4) -> min length=icl + ocl + 1
            ],
            [
                ("local", "local"),
                ("local", "sklearn"),
            ]
            + [
                ("local", "torch"),
            ]
            if TORCH_AVAILABLE
            else [],
            [False, True],
        ),
    )
    def test_with_fcm_retrain_local_and_global(self, config):
        """Test the case where the forecasting models (global and local) are retrained and the ensemble model is also
        trained. In this case, the minimum training length depends on the forecasting models' min_train_samples
        since they are retrained.
        """
        (n_points, expected_min_length), (m1_type, m2_type), multi_series = config

        m1 = NaiveSeasonal(K=4)
        # icl = K
        assert m1._target_window_lengths == (4, 0)

        if m2_type == "sklearn":
            m2 = LinearRegressionModel(lags=3, output_chunk_length=4)
        elif m2_type == "local":
            m2 = NaiveSeasonal(K=3)
            # local models don't have an output chunk length, the series can be shorter
            expected_min_length -= 4
        else:
            m2 = BlockRNNModel(
                input_chunk_length=3, output_chunk_length=4, **tfm_kwargs, n_epochs=1
            )

        self.helper_test_min_train_series_requirements(
            models=[m1, m2],
            n_points=n_points,
            train_fcm=True,
            train_with_hfc=False,
            expected_min_length=expected_min_length,
            multi_series=multi_series,
        )

    @pytest.mark.parametrize(
        "config",
        product(
            [
                (
                    2,
                    10,
                ),  # the 2 points are within first output chunk of m2 (4) -> min length=icl + ocl
                (4, 12),  # same as above
                (
                    5,
                    13,
                ),  # the fifth point is outside the first output chunk of m2 (4) -> min length=icl + ocl + 1
            ],
            [
                ("sklearn", "sklearn"),
            ]
            + [
                ("sklearn", "torch"),
                ("torch", "torch"),
            ]
            if TORCH_AVAILABLE
            else [],
            [True, False],
            [False, True],
        ),
    )
    def test_with_fcm_retrain_global_only(self, config):
        """Test the case where the forecasting models (global only) are retrained and the ensemble model is also
        trained. In this case, the minimum training length depends on the forecasting models' min_train_samples
        since they are retrained.
        """
        (
            (n_points, expected_min_length),
            (m1_type, m2_type),
            train_with_hfc,
            multi_series,
        ) = config

        if m1_type == "sklearn":
            m1 = LinearRegressionModel(lags=4, output_chunk_length=1)
        else:
            m1 = BlockRNNModel(
                input_chunk_length=4, output_chunk_length=1, **tfm_kwargs, n_epochs=1
            )

        if m2_type == "sklearn":
            m2 = LinearRegressionModel(lags=3, output_chunk_length=4)
        else:
            m2 = BlockRNNModel(
                input_chunk_length=3, output_chunk_length=4, **tfm_kwargs, n_epochs=1
            )

        self.helper_test_min_train_series_requirements(
            models=[m1, m2],
            n_points=n_points,
            train_fcm=True,
            train_with_hfc=train_with_hfc,
            expected_min_length=expected_min_length,
            multi_series=multi_series,
        )

    @pytest.mark.parametrize(
        "config",
        product(
            [
                (
                    -1,
                    8,
                ),  # -1 means at least one sample which is within first output chunk of m2 (4) -> min length=icl + ocl
                (
                    2,
                    8,
                ),  # the 2 points are within first output chunk of m2 (4) -> min length=icl + ocl
                (4, 8),  # same as above
                (
                    5,
                    9,
                ),  # the fifth point is outside the first output chunk of m2 (4) -> min length=icl + ocl + 1
            ],
            [True, False],
            [False, True],
        ),
    )
    def test_without_fcm_retrain_global_only(self, config):
        """Test the case where the forecasting models (global only) are pretrained and only the ensemble model is
        trained. In this case, the minimum training length does not depend on the forecasting models' min_train_samples
        since they are not retrained.
        """

        (n_points, expected_min_length), train_with_hfc, multi_series = config
        m1 = LinearRegressionModel(lags=4, output_chunk_length=1)
        if TORCH_AVAILABLE:
            m2 = BlockRNNModel(3, 4, **tfm_kwargs, n_epochs=1)
        else:
            m2 = LinearRegressionModel(lags=3, output_chunk_length=4)
        for model in [m1, m2]:
            model.fit(self.sine_series[: model.min_train_series_length])

        self.helper_test_min_train_series_requirements(
            models=[m1, m2],
            n_points=n_points,
            train_fcm=False,
            train_with_hfc=train_with_hfc,
            expected_min_length=expected_min_length,
            multi_series=multi_series,
        )

    def helper_test_min_train_series_requirements(
        self,
        models,
        n_points,
        train_fcm,
        train_with_hfc,
        expected_min_length,
        multi_series,
    ):
        ensemble = RegressionEnsembleModel(
            forecasting_models=models,
            regression_train_n_points=n_points,
            train_forecasting_models=train_fcm,
            train_using_historical_forecasts=train_with_hfc,
        )

        if train_fcm:
            min_samples_fc = (
                max(m_.min_train_samples for m_ in ensemble.forecasting_models) - 1
            )
        else:
            min_samples_fc = 0
        assert ensemble.min_train_series_length == expected_min_length + min_samples_fc

        # using the minimum length and `overlap_end=True` should return a single forecast
        series = self.sine_series[: ensemble.min_train_series_length]
        if multi_series:
            series = [series, series.shift(1)]
        hfc = ensemble.historical_forecasts(
            series=series,
            retrain=True,
            overlap_end=True,
            last_points_only=False,
        )
        if not multi_series:
            series = [series]
            hfc = [hfc]

        for series_, hfc_ in zip(series, hfc):
            assert len(hfc_) == 1
            assert hfc_[0].start_time() == series_.end_time() + series_.freq

        series = series[0]
        # anything less than min_train_series_length should raise an error
        with pytest.raises(ValueError) as exc:
            _ = ensemble.historical_forecasts(
                series=series[:-1],
                retrain=True,
                overlap_end=True,
                last_points_only=False,
            )
        assert str(exc.value).startswith("Cannot build any dataset to train the model")
        with pytest.raises(ValueError) as exc:
            _ = ensemble.fit(series=series[:-1])
        assert str(exc.value).startswith(
            f"`series` must have a minimum length of `{ensemble.min_train_series_length}` to fit the model."
        )

    def test_extreme_lags(self):
        # forecasting models do not use target lags
        train_n_points = 10
        model1 = RandomForestModel(
            lags_future_covariates=[0],
        )
        model2 = SKLearnModel(lags_past_covariates=3)
        model = RegressionEnsembleModel(
            forecasting_models=[model1, model2],
            regression_train_n_points=train_n_points,
        )

        assert model.extreme_lags == (None, 0, -3, -1, 0, 0, 0)
        assert model.min_train_samples == model2.min_train_samples + train_n_points

        # mix of all the lags
        model3 = RandomForestModel(
            lags_future_covariates=[-2, 5],
        )
        model4 = SKLearnModel(lags=[-7, -3], lags_past_covariates=3)
        model = RegressionEnsembleModel(
            forecasting_models=[model3, model4],
            regression_train_n_points=train_n_points,
        )

        assert model.extreme_lags == (-7, 0, -3, -1, -2, 5, 0)
        assert model.min_train_samples == model3.min_train_samples + train_n_points

    @pytest.mark.skipif(not TORCH_AVAILABLE, reason="requires torch")
    @pytest.mark.parametrize("rnn_out_larger", [False, True])
    def test_extreme_lags_torch(self, rnn_out_larger):
        # test RNN case which has the 8th extreme lags element (max_target_lag_train)
        train_n_points = 2
        icl = 20
        if rnn_out_larger:
            ocl = 1
        else:
            ocl = 20
        training_length = 24
        model = RegressionEnsembleModel(
            forecasting_models=self.get_global_models(ocl, icl, training_length),
            regression_train_n_points=train_n_points,
        )
        # RNN's training length requires more samples (training_length - icl + 1)
        # plus the points for training the regression ensemble model
        expected_samples = (training_length - icl + 1) + train_n_points
        assert model.extreme_lags == (
            -icl,
            ocl - 1,
            -icl,  # past covs from BlockRNN
            -1,  # past covs from BlockRNN
            -icl,  # future covs from RNN
            ocl - 1,  # future covs from RNN
            0,
        )
        assert model.min_train_samples == expected_samples
        assert model.min_train_series_length == icl + ocl + (expected_samples - 1)

        series = self.sine_series[: model.min_train_series_length]
        preds = model.historical_forecasts(series, overlap_end=True)
        assert preds.start_time() == series.end_time() + series.freq

        with pytest.raises(ValueError) as exc:
            model.historical_forecasts(series[:-1], overlap_end=True)
        assert str(exc.value).startswith("Cannot build any dataset to train the model")

        with pytest.raises(ValueError) as exc:
            model.fit(series[:-1])
        assert str(exc.value).startswith(
            f"`series` must have a minimum length of `{model.min_train_series_length}` to fit the model."
        )

    def test_stochastic_regression_ensemble_model(self):
        quantiles = [0.25, 0.5, 0.75]

        # probabilistic ensembling model
        linreg_prob = LinearRegressionModel(
            quantiles=quantiles, lags_future_covariates=[0], likelihood="quantile"
        )

        # deterministic ensembling model
        linreg_dete = LinearRegressionModel(lags_future_covariates=[0])

        # every models are probabilistic
        ensemble_allproba = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_probabilistic_global_model(lags=[-2, -4], quantiles=quantiles),
            ],
            regression_train_n_points=10,
            regression_model=linreg_prob.untrained_model(),
        )

        assert ensemble_allproba._models_are_probabilistic
        assert ensemble_allproba.supports_probabilistic_prediction
        ensemble_allproba.fit(self.ts_random_walk[:100])
        # probabilistic forecasting is supported
        pred = ensemble_allproba.predict(5, num_samples=10)
        assert pred.n_samples == 10

        # forecasting models are a mix of probabilistic and deterministic, probabilistic regressor
        ensemble_mixproba = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_deterministic_global_model(lags=[-2, -4]),
            ],
            regression_train_n_points=10,
            regression_model=linreg_prob.untrained_model(),
        )

        assert not ensemble_mixproba._models_are_probabilistic
        assert ensemble_mixproba.supports_probabilistic_prediction
        ensemble_mixproba.fit(self.ts_random_walk[:100])
        # probabilistic forecasting is supported
        pred = ensemble_mixproba.predict(5, num_samples=10)
        assert pred.n_samples == 10

        # forecasting models are a mix of probabilistic and deterministic, probabilistic regressor
        # with regression_train_num_samples > 1
        ensemble_mixproba2 = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_deterministic_global_model(lags=[-2, -4]),
            ],
            regression_train_n_points=10,
            regression_model=linreg_prob.untrained_model(),
            regression_train_num_samples=100,
            regression_train_samples_reduction="median",
        )

        assert not ensemble_mixproba2._models_are_probabilistic
        assert ensemble_mixproba2.supports_probabilistic_prediction
        ensemble_mixproba2.fit(self.ts_random_walk[:100])
        pred = ensemble_mixproba2.predict(5, num_samples=10)
        assert pred.n_samples == 10

        # only regression model is probabilistic
        ensemble_proba_reg = RegressionEnsembleModel(
            forecasting_models=[
                self.get_deterministic_global_model(lags=[-1, -3]),
                self.get_deterministic_global_model(lags=[-2, -4]),
            ],
            regression_train_n_points=10,
            regression_model=linreg_prob.untrained_model(),
        )

        assert not ensemble_proba_reg._models_are_probabilistic
        assert ensemble_proba_reg.supports_probabilistic_prediction
        ensemble_proba_reg.fit(self.ts_random_walk[:100])
        # probabilistic forecasting is supported
        pred = ensemble_proba_reg.predict(5, num_samples=10)
        assert pred.n_samples == 10

        # every models but regression model are probabilistics
        ensemble_dete_reg = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_probabilistic_global_model(lags=[-2, -4], quantiles=quantiles),
            ],
            regression_train_n_points=10,
            regression_model=linreg_dete.untrained_model(),
        )

        assert ensemble_dete_reg._models_are_probabilistic
        assert not ensemble_dete_reg.supports_probabilistic_prediction
        ensemble_dete_reg.fit(self.ts_random_walk[:100])
        # deterministic forecasting is supported
        ensemble_dete_reg.predict(5, num_samples=1)
        # probabilistic forecasting is not supported
        with pytest.raises(ValueError):
            ensemble_dete_reg.predict(5, num_samples=10)

        # every models are deterministic
        ensemble_alldete = RegressionEnsembleModel(
            forecasting_models=[
                self.get_deterministic_global_model([-1, -3]),
                self.get_deterministic_global_model([-2, -4]),
            ],
            regression_train_n_points=10,
            regression_model=linreg_dete.untrained_model(),
        )

        assert not ensemble_alldete._models_are_probabilistic
        assert not ensemble_alldete.supports_probabilistic_prediction
        ensemble_alldete.fit(self.ts_random_walk[:100])
        # deterministic forecasting is supported
        ensemble_alldete.predict(5, num_samples=1)
        # probabilistic forecasting is not supported
        with pytest.raises(ValueError):
            ensemble_alldete.predict(5, num_samples=10)

        # deterministic forecasters cannot be sampled
        with pytest.raises(ValueError):
            RegressionEnsembleModel(
                forecasting_models=[
                    self.get_deterministic_global_model(lags=[-1, -3]),
                    self.get_deterministic_global_model(lags=[-2, -4]),
                ],
                regression_train_n_points=10,
                regression_model=linreg_prob.untrained_model(),
                regression_train_num_samples=10,
            )

    def test_stochastic_training_regression_ensemble_model(self):
        """
        regression model is deterministic (default) but the forecasting models are
        probabilistic and they can be sampled to train the regression model.
        """
        quantiles = [0.25, 0.5, 0.75]

        # cannot sample deterministic forecasting models
        with pytest.raises(ValueError):
            RegressionEnsembleModel(
                forecasting_models=[
                    self.get_deterministic_global_model(lags=[-1, -3]),
                    self.get_deterministic_global_model(lags=[-2, -4]),
                ],
                regression_train_n_points=50,
                regression_train_num_samples=500,
            )

        # must use appropriate reduction method
        with pytest.raises(ValueError):
            RegressionEnsembleModel(
                forecasting_models=[
                    self.get_probabilistic_global_model(
                        lags=[-1, -3], quantiles=quantiles
                    ),
                    self.get_probabilistic_global_model(
                        lags=[-2, -4], quantiles=quantiles
                    ),
                ],
                regression_train_n_points=50,
                regression_train_num_samples=500,
                regression_train_samples_reduction="wrong",
            )

        # by default, does not reduce samples and convert them to components
        ensemble_model_mean = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_probabilistic_global_model(lags=[-2, -4], quantiles=quantiles),
            ],
            regression_train_n_points=50,
            regression_train_num_samples=500,
            regression_train_samples_reduction="mean",
        )

        ensemble_model_median = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_probabilistic_global_model(lags=[-2, -4], quantiles=quantiles),
            ],
            regression_train_n_points=50,
            regression_train_num_samples=500,
        )
        assert ensemble_model_median.train_samples_reduction == "median"

        ensemble_model_0_5_quantile = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_probabilistic_global_model(lags=[-2, -4], quantiles=quantiles),
            ],
            regression_train_n_points=50,
            regression_train_num_samples=500,
            regression_train_samples_reduction=0.5,
        )

        train, val = self.ts_sum1.split_after(0.9)
        ensemble_model_mean.fit(train)
        ensemble_model_median.fit(train)
        ensemble_model_0_5_quantile.fit(train)

        pred_mean_training = ensemble_model_mean.predict(len(val))
        pred_median_training = ensemble_model_median.predict(len(val))
        pred_0_5_qt_training = ensemble_model_0_5_quantile.predict(len(val))

        assert pred_median_training == pred_0_5_qt_training
        assert (
            pred_mean_training.all_values().shape
            == pred_median_training.all_values().shape
        )

        # deterministic regression model -> deterministic ensemble
        with pytest.raises(ValueError):
            ensemble_model_mean.predict(len(val), num_samples=100)
        with pytest.raises(ValueError):
            ensemble_model_median.predict(len(val), num_samples=100)
        with pytest.raises(ValueError):
            ensemble_model_0_5_quantile.predict(len(val), num_samples=100)

        # possible to use very small regression_train_num_samples
        ensemble_model_mean_1_sample = RegressionEnsembleModel(
            forecasting_models=[
                self.get_probabilistic_global_model(lags=[-1, -3], quantiles=quantiles),
                self.get_probabilistic_global_model(lags=[-2, -4], quantiles=quantiles),
            ],
            regression_train_n_points=50,
            regression_train_num_samples=1,
        )
        ensemble_model_mean_1_sample.fit(train)
        ensemble_model_mean_1_sample.predict(len(val))

        # multi-series support
        ensemble_model_median.fit([train, train + 100])
        ensemble_model_mean.predict(len(val), series=train)

    def test_predict_likelihood_parameters_univariate_regression_ensemble(self):
        quantiles = [0.05, 0.5, 0.95]
        ensemble = RegressionEnsembleModel(
            [
                self.get_probabilistic_global_model(
                    lags=2, output_chunk_length=2, quantiles=quantiles
                ),
                self.get_probabilistic_global_model(
                    lags=3, output_chunk_length=3, quantiles=quantiles
                ),
            ],
            regression_train_n_points=10,
            regression_model=LinearRegressionModel(
                lags_future_covariates=[0],
                output_chunk_length=4,
                likelihood="quantile",
                quantiles=quantiles,
            ),
        )
        ensemble.fit(self.sine_series)
        pred_ens = ensemble.predict(n=4, predict_likelihood_parameters=True)

        assert all(pred_ens.components == ["sine_q0.050", "sine_q0.500", "sine_q0.950"])
        assert all(
            pred_ens["sine_q0.050"].values() < pred_ens["sine_q0.500"].values()
        ) and all(pred_ens["sine_q0.500"].values() < pred_ens["sine_q0.950"].values())

    def test_predict_likelihood_parameters_multivariate_regression_ensemble(self):
        quantiles = [0.05, 0.5, 0.95]
        multivariate_series = self.sine_series.stack(self.lin_series)

        ensemble = RegressionEnsembleModel(
            [
                self.get_probabilistic_global_model(
                    lags=2, output_chunk_length=2, quantiles=quantiles
                ),
                self.get_probabilistic_global_model(
                    lags=3, output_chunk_length=3, quantiles=quantiles
                ),
            ],
            regression_train_n_points=10,
            regression_model=LinearRegressionModel(
                lags_future_covariates=[0],
                output_chunk_length=4,
                likelihood="quantile",
                quantiles=quantiles,
            ),
        )
        ensemble.fit(multivariate_series)
        pred_ens = ensemble.predict(n=4, predict_likelihood_parameters=True)

        assert all(
            pred_ens.components
            == [
                "sine_q0.050",
                "sine_q0.500",
                "sine_q0.950",
                "linear_q0.050",
                "linear_q0.500",
                "linear_q0.950",
            ]
        )
        assert all(
            pred_ens["sine_q0.050"].values() < pred_ens["sine_q0.500"].values()
        ) and all(pred_ens["sine_q0.500"].values() < pred_ens["sine_q0.950"].values())
        assert all(
            pred_ens["linear_q0.050"].values() < pred_ens["linear_q0.500"].values()
        ) and all(
            pred_ens["linear_q0.500"].values() < pred_ens["linear_q0.950"].values()
        )

    def test_wrong_model_creation_params(self):
        """Since `multi_models=False` requires to shift the regression model lags in the past (outside of the
        forecasting model predictions), it is not supported."""
        forcasting_models = [
            self.get_deterministic_global_model(2),
            self.get_deterministic_global_model([-5, -7]),
        ]
        RegressionEnsembleModel(
            forecasting_models=forcasting_models,
            regression_train_n_points=10,
            regression_model=LinearRegressionModel(
                lags_future_covariates=[0], output_chunk_length=2, multi_models=True
            ),
        )
        with pytest.raises(ValueError):
            RegressionEnsembleModel(
                forecasting_models=forcasting_models,
                regression_train_n_points=10,
                regression_model=LinearRegressionModel(
                    lags_future_covariates=[0],
                    output_chunk_length=2,
                    multi_models=False,
                ),
            )

    @staticmethod
    def get_probabilistic_global_model(
        lags: Union[int, list[int]],
        output_chunk_length: int = 1,
        likelihood: str = "quantile",
        quantiles: Union[None, list[float]] = [0.05, 0.5, 0.95],
        random_state: int = 42,
    ) -> LinearRegressionModel:
        return LinearRegressionModel(
            lags=lags,
            likelihood=likelihood,
            quantiles=quantiles,
            random_state=random_state,
        )

    @staticmethod
    def get_deterministic_global_model(
        lags: Union[int, list[int]], random_state: int = 13
    ) -> LinearRegressionModel:
        return LinearRegressionModel(lags=lags, random_state=random_state)



"""
TimeSeries Datasets
-------------------
"""

try:
    # inference datasets
    from darts.utils.data.torch_datasets.inference_dataset import (
        SequentialTorchInferenceDataset,
        TorchInferenceDataset,
    )

    # training datasets
    from darts.utils.data.torch_datasets.training_dataset import (
        HorizonBasedTorchTrainingDataset,
        SequentialTorchTrainingDataset,
        ShiftedTorchTrainingDataset,
        TorchTrainingDataset,
    )
except ImportError:  # Torch is not available
    from darts.utils.utils import NotImportedModule

    TorchTrainingDataset = NotImportedModule(module_name="(Py)Torch", warn=False)
    ShiftedTorchTrainingDataset = NotImportedModule(module_name="(Py)Torch", warn=False)
    SequentialTorchTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    HorizonBasedTorchTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )

    TorchInferenceDataset = NotImportedModule(module_name="(Py)Torch", warn=False)
    SequentialTorchInferenceDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )

__all__ = [
    "HorizonBasedTorchTrainingDataset",
    "TorchInferenceDataset",
    "SequentialTorchTrainingDataset",
    "TorchTrainingDataset",
    "ShiftedTorchTrainingDataset",
    "SequentialTorchInferenceDataset",
]
